mod global_state;
mod polygon;

use bevy::{
    input::{
        mouse::{MouseButtonInput, MouseWheel},
        ButtonState,
    },
    prelude::*,
    render::{
        camera::RenderTarget,
        render_resource::{
            Extent3d, TextureDescriptor, TextureDimension, TextureFormat, TextureUsages,
        },
        view::RenderLayers,
    },
    window::WindowResized,
};

use bevy_egui::{egui, EguiContexts, EguiPlugin, EguiUserTextures};
use bevy_inspector_egui::quick::WorldInspectorPlugin;
use global_state::GlobalState;
use polygon::ObjectInfos;

fn main() {
    App::new()
        .add_plugins(DefaultPlugins)
        .add_plugin(EguiPlugin)
        .add_plugin(WorldInspectorPlugin::new())
        .insert_resource(ObjectInfos::default())
        .add_startup_system(setup)
        .add_systems((mouse_event_system, update_ui_side_panel, on_resize_system))
        .run();
}

// Creates GlobalState
fn setup(
    mut cmd: Commands,
    mut images: ResMut<Assets<Image>>,
    mut egui_user_textures: ResMut<EguiUserTextures>,
) {
    let offscreen_image_size = Vec2::new(256.0, 256.0);

    let size = Extent3d {
        width: offscreen_image_size.x as u32,
        height: offscreen_image_size.y as u32,
        ..default()
    };

    // This is the texture that will be rendered to.
    let mut image = Image {
        texture_descriptor: TextureDescriptor {
            label: None,
            size,
            dimension: TextureDimension::D2,
            format: TextureFormat::Bgra8UnormSrgb,
            mip_level_count: 1,
            sample_count: 1,
            usage: TextureUsages::TEXTURE_BINDING
                | TextureUsages::COPY_DST
                | TextureUsages::RENDER_ATTACHMENT,
            view_formats: &[],
        },
        ..default()
    };

    // fill image.data with zeroes
    image.resize(size);

    let image_handle = images.add(image);
    egui_user_textures.add_image(image_handle.clone());

    // This specifies the layer used for the offscreen image, which will be attached to the first pass camera.
    let offscreen_render_layer = RenderLayers::layer(1);

    let camera_2d = cmd.spawn(Camera2dBundle::default()).id();

    let camera_3d = cmd
        .spawn((
            Camera3dBundle {
                camera: Camera {
                    target: RenderTarget::Image(image_handle.clone()),
                    ..default()
                },
                transform: Transform::from_xyz(200.0, -600.0, 600.0)
                    .looking_at(Vec3::ZERO, Vec3::Z),

                ..default()
            },
            offscreen_render_layer,
        ))
        .id();

    cmd.insert_resource(GlobalState::new(
        camera_2d,
        camera_3d,
        image_handle,
        offscreen_image_size,
        offscreen_render_layer,
    ));
}

// Create UI
fn update_ui_side_panel(
    mut contexts: EguiContexts,
    mut global_state: ResMut<GlobalState>,
    mut object_infos: ResMut<ObjectInfos>,
    mut commands: Commands,
) {
    let offscreen_texture_id = contexts.image_id(&global_state.offscreen_image).unwrap();

    egui::SidePanel::right("2D View")
        .default_width(400.0)
        .show(contexts.ctx_mut(), |ui| {
            ui.add(egui::widgets::Image::new(
                offscreen_texture_id,
                bevy_egui::egui::Vec2::new(
                    global_state.offscreen_image_size.x,
                    global_state.offscreen_image_size.y,
                ),
            ));

            if global_state.view_2d {
                if ui.button("Change to 3D View").clicked() {
                    global_state.view_2d = false;

                    // remove render layer and add camera with default render target
                    let mut cam_3d = commands.entity(global_state.camera_3d);
                    cam_3d.remove::<RenderLayers>();
                    cam_3d.remove::<Camera>();
                    cam_3d.insert(Camera::default());

                    // add render layer and add camera with offscreen render target
                    let mut cam_2d = commands.entity(global_state.camera_2d);
                    cam_2d.insert(global_state.offscreen_render_layer);
                    cam_2d.remove::<Camera>();
                    cam_2d.insert(Camera {
                        target: RenderTarget::Image(global_state.offscreen_image.clone()),
                        ..default()
                    });

                    object_infos
                        .iter_mut()
                        .for_each(|polygon| polygon.swap_to_3d(&global_state, &mut commands));
                }
            } else {
                if ui.button("Change to 2D View").clicked() {
                    global_state.view_2d = true;

                    // add render layer and add camera with offscreen render target
                    let mut cam_3d = commands.entity(global_state.camera_3d);
                    cam_3d.insert(global_state.offscreen_render_layer);
                    cam_3d.remove::<Camera>();
                    cam_3d.insert(Camera {
                        target: RenderTarget::Image(global_state.offscreen_image.clone()),
                        ..default()
                    });

                    // remove render layer and add camera with default render target
                    let mut cam_2d = commands.entity(global_state.camera_2d);
                    cam_2d.remove::<RenderLayers>();
                    cam_2d.remove::<Camera>();
                    cam_2d.insert(Camera::default());

                    object_infos
                        .iter_mut()
                        .for_each(|polygon| polygon.swap_to_2d(&global_state, &mut commands));
                }
            }

            if global_state.view_2d {
                if global_state.polygon_started {
                    // print list of coordinates of current polygon
                    if let Some(polygon) = object_infos.active_polygon() {
                        for (index, point) in polygon.points().iter().enumerate() {
                            ui.label(&format!("{}: ({}, {})", index, point.x, point.y));
                        }
                    }
                } else {
                    if ui.button("New Polygon").clicked() {
                        global_state.polygon_started = true;
                    }
                }
            }
        });
}

/// Handles all mouse events
fn mouse_event_system(
    mut mouse_button_input_events: EventReader<MouseButtonInput>,
    mut cursor_moved_events: EventReader<CursorMoved>,
    mut mouse_wheel_events: EventReader<MouseWheel>,
    mut global_state: ResMut<GlobalState>,
    mut object_infos: ResMut<ObjectInfos>,
    mut commands: Commands,
    mut meshes: ResMut<Assets<Mesh>>,
    mut color_materials: ResMut<Assets<ColorMaterial>>,
    mut standard_materials: ResMut<Assets<StandardMaterial>>,
) {
    for event in mouse_button_input_events.iter() {
        match event.button {
            MouseButton::Left => match event.state {
                ButtonState::Pressed => {
                    // only add new points if "New Polygon" button got pressed
                    if global_state.polygon_started && global_state.view_2d {
                        let polygon = match object_infos.active_polygon() {
                            Some(polygon) => polygon,
                            None => object_infos.add_polygon(),
                        };

                        if polygon.add_point(
                            &global_state,
                            &mut commands,
                            &mut meshes,
                            &mut color_materials,
                            &mut standard_materials,
                        ) {
                            global_state.polygon_started = false;
                        }
                    }
                }
                ButtonState::Released => (),
            },

            _ => (),
        }
    }

    // store latest mouse cursor location
    for event in cursor_moved_events.iter() {
        global_state.cursor_position = event.position;
    }

    for _event in mouse_wheel_events.iter() {
        //
    }
}

// Handles resize events
fn on_resize_system(
    mut resize_reader: EventReader<WindowResized>,
    mut global_state: ResMut<GlobalState>,
    mut camera: Query<&mut Transform, &Camera3d>,
) {
    for e in resize_reader.iter() {
        // When resolution is being changed
        global_state.window_extent = Vec2::new(e.width, e.height);

        let center = (global_state.window_extent * 0.5).extend(0.0);

        for mut transform in &mut camera {
            *transform = Transform::from_xyz(center.x + 200.0, center.y - 600.0, 600.0)
                .looking_at(center, Vec3::Z);
        }
    }
}
